/*[
 * GOCAD Software.
 *
 * Copyright (c) 2007-2009 Earth Decision Sciences
 * All rights reserved.
 *
 * The software and information contained herein are proprietary to, and
 * comprise valuable trade secrets of, Earth Decision Sciences, which
 * intends to preserve as trade secrets such software and information.
 * This software is furnished pursuant to a written license agreement and
 * may be used, copied, transmitted, and stored only in accordance with
 * the terms of such license and with the inclusion of the above copyright
 * notice. This software and information or any other copies thereof may
 * not be provided or otherwise made available to any other person.
]*/

#ifndef Gocad_appli_statistics_statistic_object_topology_h
#define Gocad_appli_statistics_statistic_object_topology_h

#include <Gocad/appli/enter_scope.h>

#include <Gocad/geobase/gtopo/topological_object.h>
#include <Gocad/geobase/property/property_kind.h>
#include <Gocad/math/geometry/coord.h>
#include <Gocad/math/misc/vector.h>

namespace Gocad {

/*****************************************************************************/

class _Gocad_appli_decl StatisticObjectTopoDesc{
public:
    StatisticObjectTopoDesc();
    virtual ~StatisticObjectTopoDesc();

    virtual bool is_equal( const StatisticObjectTopoDesc& other ) const = 0;
};

/*****************************************************************************/

class _Gocad_appli_decl BinStatisticObjectTopoDesc : public StatisticObjectTopoDesc {
public:
    typedef int BinIndex;
    typedef BinIndex BinCount;
public:
    BinStatisticObjectTopoDesc(
        Coord min,
        Coord max,
        CopyCString scale,
        BinCount number_of_bins
    );
    BinStatisticObjectTopoDesc(
        const BinStatisticObjectTopoDesc& from
    );
    BinStatisticObjectTopoDesc();

    virtual ~BinStatisticObjectTopoDesc();

    Coord min() const;
    Coord max() const;
    const CString& scale() const;
    BinCount number_of_bins() const;

    virtual bool is_equal( const StatisticObjectTopoDesc& other ) const;
private:
    Coord min_;
    Coord max_;
    CopyCString scale_;
    BinCount number_of_bins_;
};

/*****************************************************************************/

class _Gocad_appli_decl BinStatisticObjectTopoHandle : public TopoHandle {
public:
    typedef BinStatisticObjectTopoDesc::BinIndex BinIndex;
public:
    BinStatisticObjectTopoHandle() ;
    BinStatisticObjectTopoHandle(
        BinIndex bin_index,
        const BinStatisticObjectTopoDesc& topo_desc
    );

    const BinStatisticObjectTopoDesc& topo_desc() const;
    BinIndex bin_index() const;
    Coord bin_min() const;
    Coord bin_max() const;
    //- accessor class. DO NOT add member here.
};

/*****************************************************************************/

class _Gocad_appli_decl BinStatisticObjectTopoHandleImpl : public TopoHandleImpl {
public:
    typedef BinStatisticObjectTopoDesc::BinIndex BinIndex;
public:
    BinStatisticObjectTopoHandleImpl(
        BinIndex bin_index,
        const BinStatisticObjectTopoDesc& topo_desc
    );
    virtual ~BinStatisticObjectTopoHandleImpl();
    virtual TopoHandleImpl* copy() const;
    virtual bool copy(const TopoHandleImpl& thi);
    virtual void destroy();
private:
    friend class BinStatisticObjectTopoHandle;
    BinIndex bin_index_;
    BinStatisticObjectTopoDesc topo_desc_;
};

/*****************************************************************************/

class _Gocad_appli_decl BinStatisticObjectTopoHandleIterator : 
    public TopoHandleIterator {
public:
    typedef BinStatisticObjectTopoDesc::BinIndex BinIndex;
    typedef BinStatisticObjectTopoDesc::BinCount BinCount;

    BinStatisticObjectTopoHandleIterator(
        const BinStatisticObjectTopoDesc& topo_desc
    );
    virtual ~BinStatisticObjectTopoHandleIterator();

    virtual void reset();
    virtual Size count() const;

protected:
    void get_next_chunk();

private:
    BinIndex cur_bin_index_;
    BinStatisticObjectTopoDesc topo_desc_;
};

/*****************************************************************************/

class _Gocad_appli_decl GSBinTopologicalObject : 
    public TopologicalObjectBase
{
    GOCAD_DECLARE_RTTI( GSBinTopologicalObject );
public:
    enum { BinTopologicalDimension = 1 };

    GSBinTopologicalObject(
        const BinStatisticObjectTopoDesc& topo_desc
    );
    virtual ~GSBinTopologicalObject();
    virtual Dimension dimension() const;

    virtual TopologicalObject* child(ChildIndex) const;
    virtual ChildIndex children_count() const ;

    virtual TopoHandleIterator_var handles_itr(Dimension dim) const;
    const BinStatisticObjectTopoDesc& topo_desc() const;

    virtual Entity* owner() { return 0; }
    virtual const Entity* owner() const { return 0; }

private:
    BinStatisticObjectTopoDesc topo_desc_;
};

/*****************************************************************************/

class _Gocad_appli_decl EnumStatisticObjectTopoDesc :
    public StatisticObjectTopoDesc
{
public:
    typedef int ItemIndex;
    typedef ItemIndex ItemCount;
public:
    EnumStatisticObjectTopoDesc(
        const CString& pkind_name
    );
    EnumStatisticObjectTopoDesc(
        const EnumStatisticObjectTopoDesc& from
    );
    EnumStatisticObjectTopoDesc();

    virtual ~EnumStatisticObjectTopoDesc();

    const CString& pkind_name() const;
    const PropertyKind& pkind() const;
    const EnumPropertyKind::Mapper& mapper() const;
    ItemCount count() const;

    virtual bool is_equal( const StatisticObjectTopoDesc& other ) const;

private:
    CopyCString pkind_name_;
};

/*****************************************************************************/

class _Gocad_appli_decl EnumStatisticObjectTopoHandle : public TopoHandle {
public:
    typedef EnumStatisticObjectTopoDesc::ItemIndex ItemIndex;
    typedef DiscretePropertyKind::SignedDiscrete Code;
    typedef EnumPropertyKind::EnumElement Element;

public:
    class _Gocad_appli_decl Impl : public TopoHandleImpl {
    public:
        Impl(
            const Element& element,
            const EnumStatisticObjectTopoDesc& topo_desc
        );
        virtual ~Impl();

        virtual TopoHandleImpl* copy() const;
        virtual bool copy( const TopoHandleImpl& thi );
        virtual void destroy();

    private:
        friend class EnumStatisticObjectTopoHandle;
        Element element_;
        EnumStatisticObjectTopoDesc topo_desc_;
    };

    EnumStatisticObjectTopoHandle() ;
    EnumStatisticObjectTopoHandle(
        ItemIndex index,
        const Element& element,
        const EnumStatisticObjectTopoDesc& topo_desc
    );

    const EnumStatisticObjectTopoDesc& topo_desc() const;
    ItemIndex index() const;
    Code element_code() const;
    const CString& element_name() const;
    //- accessor class. DO NOT add member here.
};

/*****************************************************************************/

class _Gocad_appli_decl EnumStatisticObjectTopoHandleIterator : 
    public TopoHandleIterator {
public:
    typedef EnumStatisticObjectTopoDesc::ItemIndex ItemIndex;

    EnumStatisticObjectTopoHandleIterator(
        const EnumStatisticObjectTopoDesc& topo_desc
    );
    virtual ~EnumStatisticObjectTopoHandleIterator();

    virtual void reset();
    virtual Size count() const;

protected:
    void get_next_chunk();

private:
    ItemIndex cur_index_;
    EnumStatisticObjectTopoDesc topo_desc_;
    EnumPropertyKind::Mapper mapper_;
};

/*****************************************************************************/

class _Gocad_appli_decl GSEnumTopologicalObject : 
    public TopologicalObjectBase
{
    GOCAD_DECLARE_RTTI( GSEnumTopologicalObject );
public:
    enum { EnumTopologicalDimension = 0 };

    GSEnumTopologicalObject(
        const EnumStatisticObjectTopoDesc& topo_desc
    );
    virtual ~GSEnumTopologicalObject();

    virtual Dimension dimension() const;

    virtual TopologicalObject* child(ChildIndex) const;
    virtual ChildIndex children_count() const ;

    virtual TopoHandleIterator_var handles_itr( Dimension dim ) const;
    const EnumStatisticObjectTopoDesc& topo_desc() const;

    virtual Entity* owner() { return 0; }
    virtual const Entity* owner() const { return 0; }

private:
    EnumStatisticObjectTopoDesc topo_desc_;
};

/*****************************************************************************/

} // namespace Gocad

#endif
