/*[
 * GOCAD Software.
 *
 * Copyright (c) 2008 Earth Decision Sciences
 * All rights reserved.
 *
 * The software and information contained herein are proprietary to, and
 * comprise valuable trade secrets of, Earth Decision Sciences, which
 * intends to preserve as trade secrets such software and information.
 * This software is furnished pursuant to a written license agreement and
 * may be used, copied, transmitted, and stored only in accordance with
 * the terms of such license and with the inclusion of the above copyright
 * notice. This software and information or any other copies thereof may
 * not be provided or otherwise made available to any other person.
]*/

/*! \file Gocad/appli/statistics/gs_continuous_object.h
 * \brief Definition of interface class for continuous statistic objects.
 */

#include <Gocad/appli/statistics/gs_continuous_object.h>
#include <Gocad/appli/statistics/gs_statistic_object.h>

namespace Gocad {

/*****************************************************************************/

/*! \class GSContinuousBinnedObject Gocad/appli/statistics/gs_continuous_object.h
 * \brief Base class for binned statistics objects
 *
 */
GOCAD_IMPLEMENT_RTTI_0( GSContinuousBinnedObject );

/*!
 * Constructor
 */
GSContinuousBinnedObject::GSContinuousBinnedObject() {
    // Does nothing
}

/*!
 * Destructor
 */
GSContinuousBinnedObject::~GSContinuousBinnedObject()  {
}

/*!
 * Returns the associated statistic object.
 */
const GSStatisticObject* GSContinuousBinnedObject::binned_object() const {
    return dynamic_cast<const GSStatisticObject*>( this );
}

/*!
 * Returns the associated statistic object.
 */
GSStatisticObject* GSContinuousBinnedObject::binned_object() {
    return dynamic_cast<GSStatisticObject*>( this );
}

/*!
 * Returns the number of bins of the statistic object in returned topological
 * object.
 * \retval 50 by default if not defined.
 */
GSContinuousBinnedObject::BinCount GSContinuousBinnedObject::serie_nb_bins() const {
    const GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return 0; );
    int number_of_bins = 50;
    stat_obj->style()->find_attribute(
        "number_of_bins", number_of_bins
    );
    return number_of_bins;
}


/*!
 * Sets the number of bins of the histogram returned through series.
 */
void GSContinuousBinnedObject::set_serie_nb_bins( BinCount number_of_bins ) {
    GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return; );
    if( serie_nb_bins() != number_of_bins ) {
        stat_obj->style()->set_attribute(
            "number_of_bins", CopyCString::number( number_of_bins )
        );
        stat_obj->reset_topological_object();
    }
}

/*!
 * Gets the minimum abscissa of the histogram.
 * \retval 0.0 by default if it is not defined.
 */
float GSContinuousBinnedObject::serie_x_min() const {
    const GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return 0.0; );
    float min = 0;
    stat_obj->style()->find_attribute( "serie_min", min );
    return min;
}

/*!
 * Sets the minimum abscissa of the histogram.
 */
void GSContinuousBinnedObject::set_serie_x_min( float min ) {
    GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return; );
    float cur_min = serie_x_min();
    float epsilon = (min + cur_min) * 0.5e-5;
    if( !Math::equal( min, cur_min, epsilon ) ) {
        stat_obj->style()->set_attribute(
            "serie_min", CopyCString::number( min )
        );
        stat_obj->reset_topological_object();
    }
}

/*!
 * Gets the maximum abscissa of the histogram.
 * \retval serie_nb_bins()-1 by default if it is not defined.
 */
float GSContinuousBinnedObject::serie_x_max() const {
    const GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return 0.0; );
    float max = serie_nb_bins()-1;
    stat_obj->style()->find_attribute( "serie_max", max );
    return max;
}

/*!
 * Sets the maximum abscissa of the histogram.
 */
void GSContinuousBinnedObject::set_serie_x_max( float max ) {
    GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return; );
    float cur_max = serie_x_max();
    float epsilon = (max + cur_max) * 0.5e-5;
    if( !Math::equal( max, cur_max, epsilon ) ) {
        stat_obj->style()->set_attribute(
            "serie_max", CopyCString::number( max )
        );
        stat_obj->reset_topological_object();
    }
}

/*!
 * Gets the scale function of the histogram sorted values.
 * \retval "linear" by default if it is not defined.
 */
CopyCString GSContinuousBinnedObject::serie_x_scale() const {
    const GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return ""; );
    CopyCString x_scale = "linear";
    stat_obj->style()->find_attribute( "scale_function", x_scale );
    return x_scale;
}

/*! 
 * Sets the name of the scale function used to compute the step between sorted
 * values.
 */
void GSContinuousBinnedObject::set_serie_x_scale(
    const CString& x_scale
) {
    GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return; );
    CopyCString cur_scale = serie_x_scale();
    if( x_scale != cur_scale ) {
        stat_obj->style()->set_attribute( "scale_function", x_scale );
        stat_obj->reset_topological_object();
    }
}

/*!
 * Copies data from binned object \p from.
 * \param[in] from binned statistic object being copied.
 */
void GSContinuousBinnedObject::copy_binning_data(
    const GSContinuousBinnedObject& from
) {
    GSStatisticObject* stat_obj = binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( stat_obj, return; );
    const GSStatisticObject* from_stat_obj = from.binned_object();
    // All continuous object should be statistic object.
    GOCAD_ASSERT_POINTER( from_stat_obj, return; );
    CopyCStringList attributes(4);
    attributes.append("number_of_bins");
    attributes.append("serie_max");
    attributes.append("serie_min");
    attributes.append("scale_function");
    for ( CopyCStringList::iterator itr(attributes);itr.more();itr.next() ) {
        CopyCString attribute = itr.cur();
        CopyCString attribute_value;
        if(
            from_stat_obj->style()->find_attribute(
                attribute, attribute_value
            )
        ) {
            stat_obj->style()->set_attribute(
                attribute, attribute_value
            );
        }
    }
}

/*****************************************************************************/

/*! \class GSStatisticObjectDiscreteModelDescriptor Gocad/appli/statistics/gs_continuous_object.h
 * \ingroup DataAnalysis
 * \brief Base abstract class for all discrete model based on bins
 * descriptors.
 */

GOCAD_IMPLEMENT_RTTI_1(
    GSStatisticObjectDiscreteModelDescriptor,
    GSStatisticObjectModelDescriptor
);

/*! \typedef int GSStatisticObjectDiscreteModelDescriptor::Index
 * Index of the associated model. The index is used to distinguish two discrete
 * models of the same object based on the same method (DSI, Kernel...).
 * It should be used only for the descriptor pretty user name, two descriptors
 * having the same parameters but the index as considered as equal.
 */

/*!
 * Constructor. The default index is -1. The default scale is "linear".
 */
GSStatisticObjectDiscreteModelDescriptor::GSStatisticObjectDiscreteModelDescriptor() :
    index_( -1 ),
    scale_type_( "linear" )
{
    // Does nothing
}

/*!
 * Default destructor.
 */
GSStatisticObjectDiscreteModelDescriptor::~GSStatisticObjectDiscreteModelDescriptor() {
    // Does nothing
}

/*!
 * Sets the index of the model.
 */
void GSStatisticObjectDiscreteModelDescriptor::set_index( Index index ) {
    index_ = index;
}

/*!
 * Returns the index of the model.
 */
GSStatisticObjectDiscreteModelDescriptor::Index
GSStatisticObjectDiscreteModelDescriptor::index() const {
    return index_;
}

/*!
 * Defines the scale type that is associated to the discrete model. Discrete
 * models are bound to one scale ("linear", "log10"...) and should not be
 * used in space with different scale.
 */
void GSStatisticObjectDiscreteModelDescriptor::set_scale_type(
    const CString& scale_type
) {
    scale_type_ = scale_type;
}

/*!
 * Returns the associated scale type.
 */
const CString& GSStatisticObjectDiscreteModelDescriptor::scale_type() const {
    return scale_type_;
}

/*!
 * \copydoc GSStatisticObjectModel::Descriptor::is_equal
 * Checks that the given descriptor is defined with the same scale.
 */
bool GSStatisticObjectDiscreteModelDescriptor::is_equal(
    const GSStatisticObject::DescriptorBase& other
) const {
    const GSStatisticObjectDiscreteModelDescriptor* odesc =
        dynamic_cast<const GSStatisticObjectDiscreteModelDescriptor*>(
            &other
        );
    if( odesc == nil ) {
        return false;
    }
    if( odesc->scale_type_ != scale_type_ ) {
        return false;
    }
    return GSStatisticObjectModelDescriptor::is_equal( other );
}

/*!
 * Returns "Model#index".
 */
Format GSStatisticObjectDiscreteModelDescriptor::base_name() const {
    return Format::tr( "Model#%1" ) % index();
}

/*!
 * Creates a new GSStatisticObjectDiscreteModelDescriptor that is a copy of this.
 */
GSStatisticObject::DescriptorBase*
GSStatisticObjectDiscreteModelDescriptor::clone() const {
    GSStatisticObjectDiscreteModelDescriptor* copy = new
        GSStatisticObjectDiscreteModelDescriptor();

    // Copies the GSStatisticObjectModel::Descriptor data
    copy_data( *copy );
    return copy;
}

/*!
 * Copies index and scale type.
 */
void GSStatisticObjectDiscreteModelDescriptor::copy_data(
    GSStatisticObjectModelDescriptor& to
) const {
    GSStatisticObjectDiscreteModelDescriptor* to_desc =
        dynamic_cast<GSStatisticObjectDiscreteModelDescriptor*>( &to );
    if( to_desc != nil ) {
        to_desc->index_ = index_;
        to_desc->scale_type_ = scale_type_;
    }
    GSStatisticObjectModelDescriptor::copy_data( to );
}

/*****************************************************************************/

} // namespace Gocad
