/*[
 * GOCAD Software.
 *
 * Copyright (c) 2008 Earth Decision Sciences
 * All rights reserved.
 *
 * The software and information contained herein are proprietary to, and
 * comprise valuable trade secrets of, Earth Decision Sciences, which
 * intends to preserve as trade secrets such software and information.
 * This software is furnished pursuant to a written license agreement and
 * may be used, copied, transmitted, and stored only in accordance with
 * the terms of such license and with the inclusion of the above copyright
 * notice. This software and information or any other copies thereof may
 * not be provided or otherwise made available to any other person.
]*/

/*! \file Gocad/appli/statistics/statistic_object_topology.h
 * \brief Classes related to topology for statistic objects
 *
 * Main class is GSBinTopologicalObject
 */


/* TRANSLATOR Gocad::Format */

#include <Gocad/appli/statistics/statistic_object_topology.h>
#include <Gocad/utils/os/breakpoint.h>

namespace Gocad {

/*****************************************************************************/
/*! \class StatisticObjectTopoDesc Gocad/appli/statistics/statistic_object_topology.h
 * \brief Base class for statistical object topological description
 */

/*!
 * Constructor
 */
StatisticObjectTopoDesc::StatisticObjectTopoDesc() {
}

/*!
 * Destructor
 */
StatisticObjectTopoDesc::~StatisticObjectTopoDesc() {
}

/*****************************************************************************/
/*! \class BinStatisticObjectTopoDesc Gocad/appli/statistics/statistic_object_topology.h
 * \brief Base class for binned statistical object topological description
 */


/*! \typedef int BinStatisticObjectTopoDesc::BinIndex
 *
 */

    
/*! \typedef BinStatisticObjectTopoDesc::BinIndex BinStatisticObjectTopoDesc::BinCount
 *
 */

/*!
 * Constructor
 */
BinStatisticObjectTopoDesc::BinStatisticObjectTopoDesc() :
    min_( 0.0 ),
    max_( 0.0 ),
    scale_( "linear" ),
    number_of_bins_( 0 )
{
    // Does nothing
}

/*!
 * Constructor
 * \param[in] min starting value for the first bin.
 * \param[in] max end value for the last bin.
 * \param[in] scale scale used to fill the bins.
 * \param[in] number_of_bins number of bins.
 */

BinStatisticObjectTopoDesc::BinStatisticObjectTopoDesc(
    Coord min,
    Coord max,
    CopyCString scale,
    BinCount number_of_bins
) : min_(min),
    max_(max),
    scale_(scale),
    number_of_bins_(number_of_bins) {
}

/*!
 * Copy constructor
 */
BinStatisticObjectTopoDesc::BinStatisticObjectTopoDesc(
    const BinStatisticObjectTopoDesc& from
) : min_(from.min_),
    max_(from.max_),
    scale_(from.scale_),
    number_of_bins_(from.number_of_bins_) {
}

/*!
 * Destructor
 */
BinStatisticObjectTopoDesc::~BinStatisticObjectTopoDesc() {
}

/*!
 * \return the number of bins.
 */
BinStatisticObjectTopoDesc::BinCount 
BinStatisticObjectTopoDesc::number_of_bins() const {
    return number_of_bins_;
}

/*!
 * \return the starting value for the first bin.
 */
float BinStatisticObjectTopoDesc::min() const {
    return min_;
}

/*!
 * \return the end value for the last bin.
 */
float BinStatisticObjectTopoDesc::max() const {
    return max_;
}

/*!
 * \return the scale used to fill the bins.
 */
const CString& BinStatisticObjectTopoDesc::scale() const {
    return scale_;
}

/*!
 * \return true if the given object is equal to this.
 */
bool BinStatisticObjectTopoDesc::is_equal(
    const StatisticObjectTopoDesc& desc
) const {
    const BinStatisticObjectTopoDesc* desc_ptr =
        dynamic_cast<const BinStatisticObjectTopoDesc*>( &desc );
    if ( desc_ptr == nil ) {
        return false;
    }
    if ( min_ != desc_ptr->min_ 
        || max_ != desc_ptr->max_
        || scale_ != desc_ptr->scale_
        || number_of_bins_ != desc_ptr->number_of_bins_ ) {
        return false;
    }

    return true;
}

/*****************************************************************************/
/*! \class BinStatisticObjectTopoHandle Gocad/appli/statistics/statistic_object_topology.h
 * \brief Accessor class to BinStatisticObjectTopoHandleImpl 
 * which contains topological information of a statistical object
 *
 * \sa BinStatisticObjectTopoHandleImpl
 */

/*! \typedef  BinStatisticObjectTopoDesc::BinIndex BinStatisticObjectTopoHandle::BinIndex
 *
 */


/*!
 * Empty constructor
 */
BinStatisticObjectTopoHandle::BinStatisticObjectTopoHandle() : TopoHandle() {
}

/*!
 * Constructor
 * \param[in] bin_index index of the bin.must be between 0 and number_of_bins
 * \param[in] topo_desc descriptor of needed topological information, min
 * max, scale  and number of bins of the binned object.
 */
BinStatisticObjectTopoHandle::BinStatisticObjectTopoHandle(
    BinIndex bin_index,
    const BinStatisticObjectTopoDesc& topo_desc
) : TopoHandle() {
    impl_ = new BinStatisticObjectTopoHandleImpl(
        bin_index,
        topo_desc
    );
    h_.index = static_cast<BinStatisticObjectTopoHandleImpl*>(impl_)->bin_index_;
}

/*!
 * \return the index of the bin.
 */
BinStatisticObjectTopoHandle::BinIndex 
BinStatisticObjectTopoHandle::bin_index() const {
    if (impl_ == nil) {
        return 0.;
    }
    return static_cast<BinStatisticObjectTopoHandleImpl*>(impl_)->bin_index_;
}

/*!
 * \return the topology descriptor.
 */
const BinStatisticObjectTopoDesc& 
BinStatisticObjectTopoHandle::topo_desc() const {
    if (impl_ == nil) {
        static BinStatisticObjectTopoDesc null_topo_desc;
        return null_topo_desc;
    }
    return static_cast<BinStatisticObjectTopoHandleImpl*>(impl_)->topo_desc_;
}

/*!
 * Return the minimum value of the bin.
 */
Coord BinStatisticObjectTopoHandle::bin_min() const {
    const BinStatisticObjectTopoDesc& desc = topo_desc();
    if( desc.number_of_bins() == 0 ) {
        return 0.0;
    }
    return desc.min() +
        Coord( bin_index() ) * ( desc.max() - desc.min() ) /
        Coord( desc.number_of_bins() );
}

/*!
 * Return the maximum value of the bin.
 */
Coord BinStatisticObjectTopoHandle::bin_max() const {
    const BinStatisticObjectTopoDesc& desc = topo_desc();
    if( desc.number_of_bins() == 0 ) {
        return 0.0;
    }
    return desc.min() +
        Coord( bin_index() + 1.0 ) * ( desc.max() - desc.min() ) /
        Coord( desc.number_of_bins() );
}

/*****************************************************************************/
/*! \class BinStatisticObjectTopoHandleImpl Gocad/appli/statistics/statistic_object_topology.h
 * \brief contains topological information of a statistical object
 *
 */

/*! \typedef  BinStatisticObjectTopoDesc::BinIndex BinStatisticObjectTopoHandleImpl::BinIndex
 *
 */


/*!
 * Constructor
 * \param[in] bin_index index of the bin.
 * \param[in] topo_desc descriptor of needed topological information, min
 * max, scale  and number of bins of the binned object.
 */
BinStatisticObjectTopoHandleImpl::BinStatisticObjectTopoHandleImpl(
    BinIndex bin_index,
    const BinStatisticObjectTopoDesc& topo_desc
) : bin_index_(bin_index),
    topo_desc_(topo_desc) {
}

/*!
 * Destructor
 */
BinStatisticObjectTopoHandleImpl::~BinStatisticObjectTopoHandleImpl() {
}

/*!
 * Creates a copy of this BinStatisticObjectTopoHandleImpl
 */
TopoHandleImpl* BinStatisticObjectTopoHandleImpl::copy() const {
    return new BinStatisticObjectTopoHandleImpl(
        bin_index_,
        topo_desc_
    );
}

/*!
 * Copy constructor.
 */
bool BinStatisticObjectTopoHandleImpl::copy(const TopoHandleImpl& thi) {
    const BinStatisticObjectTopoHandleImpl& sothi = 
        static_cast<const BinStatisticObjectTopoHandleImpl&>(thi);
    bin_index_ = sothi.bin_index_;
    topo_desc_ = sothi.topo_desc_;
    return true;
}

/*!
 * delete this.
 */
void BinStatisticObjectTopoHandleImpl::destroy() {
    delete this;
}


/*****************************************************************************/
/*! \class BinStatisticObjectTopoHandleIterator Gocad/appli/statistics/statistic_object_topology.h
 * \brief Iterates through all topohandle of a BinStatisticObject.
 *
 * BinStatisticObjectTopoHandleIterator iterates through all topohandle (bins)
 * of a BinStatisticObject.
 */

/*!
 * \typedef BinStatisticObjectTopoDesc::BinIndex BinStatisticObjectTopoHandleIterator::BinIndex
 */

/*!
 * \typedef BinStatisticObjectTopoDesc::BinCount BinStatisticObjectTopoHandleIterator::BinCount
 */


/*!
 * BinStatisticObjectTopoHandleIterator iterates through all topohandle (bins)
 * of a BinStatisticObject.
 */
BinStatisticObjectTopoHandleIterator::BinStatisticObjectTopoHandleIterator(
    const BinStatisticObjectTopoDesc& topo_desc
) : TopoHandleIterator(),
    topo_desc_(topo_desc),
    cur_bin_index_(0)
{
    get_next_chunk();
}


/*!
 * Destructor.
 */
BinStatisticObjectTopoHandleIterator::~BinStatisticObjectTopoHandleIterator() {
}

/*!
 * Reinitialize the iterator.
 */
void BinStatisticObjectTopoHandleIterator::reset() {
    cur_bin_index_ = 0;
    get_next_chunk();
}

/*!
 * \copydoc ChunkIterator::get_next_chunk()
 */
void BinStatisticObjectTopoHandleIterator::get_next_chunk() {
    int index = -1;
    int max_bin_index = chunk_size()-1;
    for( ; 
        cur_bin_index_<topo_desc_.number_of_bins() && index<max_bin_index;
        cur_bin_index_++
    ) {
        index++;
        BinStatisticObjectTopoHandle& goth = 
            static_cast<BinStatisticObjectTopoHandle&>(chunk()[index]);
        goth = BinStatisticObjectTopoHandle(
            cur_bin_index_,
            topo_desc_
        );
    }
    set_index(index);
}

/*!
 * \return the number of bins.
 */
BinStatisticObjectTopoHandleIterator::Size BinStatisticObjectTopoHandleIterator::count() const {
    return topo_desc_.number_of_bins();
}

/*****************************************************************************/
/*! \class GSBinTopologicalObject Gocad/appli/statistics/statistic_object_topology.h
 * \brief Topological object associated to a BinStatisticObject.
 *
 * Topological object associated to a BinStatisticObject. 
 * It is defined by its number of bins,
 * minimum value for the first bin, a maximum value f
 * or the last bin and the scale used to fill the bin.
 */

GOCAD_IMPLEMENT_RTTI_1( GSBinTopologicalObject,TopologicalObjectBase );

/*!
 * Constructor
 *
 * \param[in] topo_desc_ is the topology descriptor.
 */
GSBinTopologicalObject::GSBinTopologicalObject(
    const BinStatisticObjectTopoDesc& topo_desc
) : topo_desc_(topo_desc) {
}

/*!
 * Destructor
 */
GSBinTopologicalObject::~GSBinTopologicalObject() {
}

/*!
 * \copydoc TopologicalObject::dimension()
 */ 
TopologicalObject::Dimension 
GSBinTopologicalObject::dimension() const {
    return 1;
}


/*!
 * \copydoc TopologicalObject::child(TopologicalObject::ChildIndex)
 */ 
TopologicalObject* GSBinTopologicalObject::child(
    TopologicalObject::ChildIndex
) const {
    return nil;
}

/*!
 * \copydoc TopologicalObject::children_count()
 */ 
TopologicalObject::ChildIndex
GSBinTopologicalObject::children_count() const {
    return 0;
}

/*!
 * \copydoc TopologicalObject::handles_itr(Dimension)
 */ 
TopoHandleIterator_var GSBinTopologicalObject::handles_itr(
    Dimension
) const {
    return new BinStatisticObjectTopoHandleIterator(topo_desc_);
}

/*!
 * \return topological descriptor
 */
const BinStatisticObjectTopoDesc& 
GSBinTopologicalObject::topo_desc() const {
    return topo_desc_;
}


/*****************************************************************************/

/*! \class EnumStatisticObjectTopoDesc Gocad/appli/statistics/statistic_object_topology.h
 * \brief Base class for categorical statistical object topological
 * description.
 *
 * The topology of categorical objects is dynamic and may be changed when
 * elements are re-ordered in a given property kind.
 */

/*! \typedef int EnumStatisticObjectTopoDesc::ItemIndex
 * Index of current element.
 */

/* \typedef EnumStatisticObjectTopoDesc::ItemIndex EnumStatisticObjectTopoDesc::ItemCount
 * Number of elements.
 */

/*!
 * Default constructor.
 */
EnumStatisticObjectTopoDesc::EnumStatisticObjectTopoDesc() :
    pkind_name_()
{
    // Does nothing
}

/*!
 * Constructor. The given string should be the name of an existing
 * EnumPropertyKind.
 */
EnumStatisticObjectTopoDesc::EnumStatisticObjectTopoDesc(
    const CString& pkind_name
) :
    pkind_name_( pkind_name )
{
    // Does nothing
}

/*!
 * Copy constructor.
 */
EnumStatisticObjectTopoDesc::EnumStatisticObjectTopoDesc(
    const EnumStatisticObjectTopoDesc& from
) :
    pkind_name_( from.pkind_name_ )
{
    // Does nothing
}

/*!
 * Destructor
 */
EnumStatisticObjectTopoDesc::~EnumStatisticObjectTopoDesc() {
}

/*!
 * Returns the PropertyKind name.
 */
const CString& EnumStatisticObjectTopoDesc::pkind_name() const {
    return pkind_name_;
}

/*!
 * Returns the associated PropertyKind.
 */
const PropertyKind& EnumStatisticObjectTopoDesc::pkind() const {
    return PropertyKindCatalog::instance()->find( pkind_name_ );
}

/*!
 * Returns the current mapper of the associated PropertyKind.
 * \retval An empty mapper if the associated PropertyKind is not an
 * EnumPropertyKind.
 */
const EnumPropertyKind::Mapper& EnumStatisticObjectTopoDesc::mapper() const {
    const EnumPropertyKind* epk =
        dynamic_cast<const EnumPropertyKind*>( &pkind() );
    if( epk == nil ) {
        static EnumPropertyKind::Mapper empty_mapper;
        return empty_mapper;
    }
    return epk->mapper();
}

/*!
 * Returns the number of items defined in the mapper.
 */
EnumStatisticObjectTopoDesc::ItemCount
EnumStatisticObjectTopoDesc::count() const {
    return mapper().count();
}

/*!
 * Tests whether the given object is equal to this or not.
 */
bool EnumStatisticObjectTopoDesc::is_equal(
    const StatisticObjectTopoDesc& desc
) const {
    const EnumStatisticObjectTopoDesc* desc_ptr =
        dynamic_cast<const EnumStatisticObjectTopoDesc*>( &desc );
    if ( desc_ptr == nil ) {
        return false;
    }
    if( pkind_name_ != desc_ptr->pkind_name_ ) {
        return false;
    }

    return true;
}

/*****************************************************************************/

/*! \class EnumStatisticObjectTopoHandle::Impl Gocad/appli/statistics/statistic_object_topology.h
 * \brief contains topological information of a categorical statistical object
 */

/*!
 * Constructor.
 * \param[in] element Element.
 * \param[in] topo_desc Descriptor of needed topological information
 */
EnumStatisticObjectTopoHandle::Impl::Impl(
    const Element& element,
    const EnumStatisticObjectTopoDesc& topo_desc
) :
    element_( element ),
    topo_desc_( topo_desc )
{
    // Does nothing
}

/*!
 * Destructor
 */
EnumStatisticObjectTopoHandle::Impl::~Impl() {
    // Does nothing
}

/*!
 * Creates a copy of this EnumStatisticObjectTopoHandle::Impl
 */
TopoHandleImpl* EnumStatisticObjectTopoHandle::Impl::copy() const {
    return new Impl( element_, topo_desc_ );
}

/*!
 * Copy constructor.
 */
bool EnumStatisticObjectTopoHandle::Impl::copy( const TopoHandleImpl& thi ) {
    const Impl& sothi =  static_cast<const Impl&>( thi );
    element_ = sothi.element_;
    topo_desc_ = sothi.topo_desc_;
    return true;
}

/*!
 * Deletes this.
 */
void EnumStatisticObjectTopoHandle::Impl::destroy() {
    delete this;
}

/*****************************************************************************/

/*! \class EnumStatisticObjectTopoHandle Gocad/appli/statistics/statistic_object_topology.h
 * \brief Accessor class to EnumStatisticObjectTopoHandleImpl 
 * which contains topological information of a categorical statistical object
 *
 * \sa EnumStatisticObjectTopoHandle::Impl
 */

/*! \typedef EnumStatisticObjectTopoDesc::ItemIndex EnumStatisticObjectTopoHandle::ItemIndex
 * Index of current element.
 */

/* \typedef DiscretePropertyKind::SignedDiscrete EnumStatisticObjectTopoHandle::Code
 * Code of the current element.
 */

/* \typedef EnumPropertyKind::EnumElement EnumStatisticObjectTopoHandle::Element
 * Current element.
 */

/*!
 * Empty constructor
 */
EnumStatisticObjectTopoHandle::EnumStatisticObjectTopoHandle() : TopoHandle() {
    // Does nothing
}

/*!
 * Constructor
 * \param[in] index index of the element. Must be between 0 and
 * EnumStatisticObjectTopoDesc::count().
 * \param[in] element Element.
 * \param[in] topo_desc descriptor of needed topological information.
 */
EnumStatisticObjectTopoHandle::EnumStatisticObjectTopoHandle(
    ItemIndex index,
    const Element& element,
    const EnumStatisticObjectTopoDesc& topo_desc
) :
    TopoHandle( index )
{
    impl_ = new Impl( element, topo_desc );
}

/*!
 * Returns the topology descriptor.
 *
 * \warning The mapper returned by the descriptor may be not synchronized with
 * this index.
 * \sa EnumStatisticObjectTopoHandleIterator
 */
const EnumStatisticObjectTopoDesc& 
EnumStatisticObjectTopoHandle::topo_desc() const {
    if( impl_ == nil ) {
        static EnumStatisticObjectTopoDesc null_topo_desc;
        return null_topo_desc;
    }
    return static_cast<Impl*>( impl_ )->topo_desc_;
}

/*!
 * Returns the index of the element.
 */
EnumStatisticObjectTopoHandle::ItemIndex 
EnumStatisticObjectTopoHandle::index() const {
    return h_.index;
}

/*!
 * Returns the code of the element.
 * \retval -1 if no element is set.
 */
EnumStatisticObjectTopoHandle::Code
EnumStatisticObjectTopoHandle::element_code() const {
    if( impl_ == nil ) {
        return -1;
    }
    return static_cast<Impl*>( impl_ )->element_.index;
}

/*!
 * Returns the name of the element.
 * \retval CString::null() if no element is set.
 */
const CString& EnumStatisticObjectTopoHandle::element_name() const {
    if( impl_ == nil ) {
        return CString::null();
    }
    return static_cast<Impl*>( impl_ )->element_.name;
}

/*****************************************************************************/

/*! \class EnumStatisticObjectTopoHandleIterator Gocad/appli/statistics/statistic_object_topology.h
 * \brief Iterates through all topohandle of a GSEnumTopologicalObject.
 *
 * EnumStatisticObjectTopoHandleIterator iterates through all topohandles
 * (discrete elements) of a GSEnumTopologicalObject.
 */

/*! \typedef EnumStatisticObjectTopoDesc::ItemIndex EnumStatisticObjectTopoHandleIterator::ItemIndex
 * Index of current element.
 */

/*!
 * EnumStatisticObjectTopoHandleIterator iterates through all topohandles
 * (discrete elements) of a GSEnumTopologicalObject.
 *
 * The order of the elements is the one defined in the PropertyKind at the
 * instantiation time or at the time reset() is called.
 */
EnumStatisticObjectTopoHandleIterator::
EnumStatisticObjectTopoHandleIterator(
    const EnumStatisticObjectTopoDesc& topo_desc
) :
    TopoHandleIterator(),
    topo_desc_( topo_desc ),
    cur_index_( 0 )
{
    mapper_ = topo_desc_.mapper();
    get_next_chunk();
}

/*!
 * Destructor.
 */
EnumStatisticObjectTopoHandleIterator::
~EnumStatisticObjectTopoHandleIterator() {
    // Does nothing
}

/*!
 * Reinitializes the iterator.
 */
void EnumStatisticObjectTopoHandleIterator::reset() {
    cur_index_ = 0;
    mapper_ = topo_desc_.mapper();
    get_next_chunk();
}

/*!
 * \copydoc ChunkIterator::get_next_chunk()
 */
void EnumStatisticObjectTopoHandleIterator::get_next_chunk() {
    int index = -1;
    int max_bin_index = chunk_size()-1;
    for(
        ; cur_index_ < mapper_.count() && index < max_bin_index;
        cur_index_++
    ) {
        index++;
        EnumStatisticObjectTopoHandle& goth = 
            static_cast<EnumStatisticObjectTopoHandle&>( chunk()[index] );
        goth = EnumStatisticObjectTopoHandle(
            cur_index_,
            mapper_.item( cur_index_ ),
            topo_desc_
        );
    }
    set_index( index );
}

/*!
 * Returns the number of elements.
 */
EnumStatisticObjectTopoHandleIterator::Size
EnumStatisticObjectTopoHandleIterator::count() const {
    return mapper_.count();
}

/*****************************************************************************/

/*! \class GSEnumTopologicalObject Gocad/appli/statistics/statistic_object_topology.h
 * \brief Topological object associated to a categorical statistic object.
 *
 * Topological object associated to a categorical statistic object. 
 * It is defined by a property kind that provides which elements and in which
 * order are used to define the topology.
 */

GOCAD_IMPLEMENT_RTTI_1( GSEnumTopologicalObject, TopologicalObjectBase );

/*!
 * Constructor
 *
 * \param[in] topo_desc is the topology descriptor.
 */
GSEnumTopologicalObject::GSEnumTopologicalObject(
    const EnumStatisticObjectTopoDesc& topo_desc
) :
    topo_desc_( topo_desc )
{
    // Does nothing
}

/*!
 * Destructor
 */
GSEnumTopologicalObject::~GSEnumTopologicalObject() {
    // Does nothing
}

/*!
 * \copydoc TopologicalObject::dimension()
 */ 
TopologicalObject::Dimension 
GSEnumTopologicalObject::dimension() const {
    return EnumTopologicalDimension;
}

/*!
 * \copydoc TopologicalObject::child( TopologicalObject::ChildIndex )
 */ 
TopologicalObject* GSEnumTopologicalObject::child(
    TopologicalObject::ChildIndex
) const {
    return nil;
}

/*!
 * \copydoc TopologicalObject::children_count()
 */ 
TopologicalObject::ChildIndex
GSEnumTopologicalObject::children_count() const {
    return 0;
}

/*!
 * \copydoc TopologicalObject::handles_itr(Dimension)
 */ 
TopoHandleIterator_var GSEnumTopologicalObject::handles_itr(
    Dimension
) const {
    return new EnumStatisticObjectTopoHandleIterator( topo_desc_ );
}

/*!
 * Returns topological descriptor.
 */
const EnumStatisticObjectTopoDesc& GSEnumTopologicalObject::topo_desc() const {
    return topo_desc_;
}

/*****************************************************************************/

} // namespace Gocad
